/**
 * \file LRU_queue.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief The header for the LRU_queue class.
 *
 * This file contains the header that defines an LRU_queue object,
 * which is capable of keeping pages in an LRU order based on a
 * sequence of references.
 **/



// ===================================================================
#if !defined (_LRU_QUEUE_HH)
#define _LRU_QUEUE_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <hash_map.h>
#include <list.h>
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// TYPES

/** \brief The type for a queue of page_s objects. */
typedef list<page_s> queue_t;

/**
 * \brief The type for a table that maps page_s's to queue_t
 *        positions.
 **/
typedef hash_map<page_s,
		 queue_t::iterator,
		 page_hash_s,
		 equal_pages_s> queue_position_map_t;
// ===================================================================



// ===================================================================
/**
 * \brief A queue of page_s objects maintained in LRU order.
 *
 *  An LRU_queue allows for an arbitrary sequence of page uses, each
 *  time returning information as to where in the queue ordering the
 *  page was found (if at all) and which page was evicted from the
 *  given memory size (if one was).
 **/
class LRU_queue {

protected: // DATA MEMBERS

  unsigned int capacity; /**< The number of pages the queue can
			    hold. */
  unsigned int size; /**< The current number of pages held by the
			queue. */
  queue_t queue; /**< The underlying queue that orders the page_ss. */

  queue_position_map_t queue_position_map; /**< A map from page ID to
					      position in the
					      queue. */

public: // CONSTANT DATA MEMBERS

  /**
   * \brief A special page_s that indicates a lack of eviction.
   * \see LRU_queue::LRU_queue()
   *
   * Defined using constants that guarantee that no real page could
   * have this identity.  It sets the highest order bit, which cannot
   * be part of a legitimate page number.
   **/
  const page_s no_evict;

public: // METHODS

  /**
   * \brief The constructor.
   * \param init_capacity The number of pages the queue can store.
   *
   * Create an LRU_queue which will order pages according to their
   * sequence of reference.  Initialize the data structures used to
   * order and communicate the results of each reference.
   **/
  LRU_queue (const unsigned int init_capacity) : 
    capacity(init_capacity),
    size(0),
    queue(),
    queue_position_map(),
    no_evict(0, 1 << ((sizeof(page_ID_t) * BITS_PER_BYTE) - 1)) {}

  /**
   * \brief Whether the queue is empty.
   * \return True for an empty queue; false otherwise.
   **/
  bool empty () {
    return (size == 0);
  }

  /**
   * \brief Whether the queue is full.
   * \return True for a full queue; false otherwise.
   **/
  bool full () {
    return (size == capacity);
  }

  /**
   * \brief Reference a particular page.
   * \param page A pointer to the page to be referenced.
   * \return The page that was evicted, or the special page
   *         <tt>no_evict</tt> if no page was evicted.
   * \see no_evict
   *
   * Attempt to find a particular page in the queue.  If it is found,
   * move it to the front of the queue, and return the lack of
   * eviction from the queue.  If it was not found, insert it newly
   * into the queue and return the page that was evicted from the LRU
   * end of the queue (if any).
   **/
  page_s touch (const page_s* const page);

protected: // METHODS

  /**
   * \brief Choose a page for eviction from the end of the queue.
   * \return The page chosen for eviction.
   **/
  page_s evict () {
    page_s page = *(--queue.end());
    queue.pop_back();
    queue_position_map[page] = queue.end();
    size--;
    return page;
  }

  /**
   * \brief Insert a page at the front of the queue.
   * \param page The page to be inserted.
   **/
  void push_node_on_front (page_s page) {
    queue.push_front(page);
    queue_position_map[page] = queue.begin();
  }

  /**
   * \return The current most recently used page.
   **/
  page_s top () {
    return *(queue.begin());
  }

  /**
   * \brief Respond to a reference to a page not in the queue.
   * \param page A pointer to the referenced page that caused the
   *        miss.
   * \return The page evicted as a result of the miss, or
   *         <tt>no_evict</tt> if the queue was not full.
   * \see no_evict
   *
   * When a reference has been identified as a miss, insert the
   * referenced page at the front of the queue and select a page for
   * eviction if the queue is full.
   **/
  page_s handle_miss (const page_s* const page) {
    page_s evicted = no_evict;
    if (full()) evicted = evict();
    push_node_on_front(*page);
    size++;
    return evicted;
  }
  
  /**
   * \brief Respond to a reference to a page already in the queue.
   * \param page A pointer to the referenced page that caused the
   *        hit.
   * \param i An iterator to the queue position that holds the
   *        referenced page.
   *
   * Move the referenced page to the front of the queue and return the
   * position at which it was found.
   **/
  void handle_hit (const page_s* const page, queue_t::iterator i) {

    queue.erase(i);
    push_node_on_front(*page);

  }

}; // LRU_queue
// ===================================================================



// ===================================================================
#endif // _LRU_QUEUE_HH
// ===================================================================
