/**
 * \file SAD_filter.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu> and Yannis
 *         Smaragdakis <yannis@cc.gatech.edu>
 * \date January 2003
 * \brief The code for the SAD class which performs trace reduction.
 *
 * The methods of a SAD filter, which reduces a reference stream and
 * emits the results.
 **/



// ===================================================================
// INCLUDES

#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h> // For S_IRUSR and S_IWUSR
#include "SAD_filter.hh"
// ===================================================================



// ===================================================================
SAD_filter::SAD_filter (const unsigned int reduction_memory_size,
			const char* const output_directory_pathname,
			const task_ID_t init_task_ID,
			FILE* const init_task_ID_list_stream) :
  task_ID(init_task_ID),
  task_ID_list_stream(init_task_ID_list_stream),
  lru_queue(reduction_memory_size) {

  // Construct the output pathname for later use and nullify the output
  // stream.  (The output file is opened only when the first record is
  // added to the filter.)
  sprintf(output_pathname,
	  "%s/%hd.sad-%d.gz",
	  output_directory_pathname,
	  task_ID,
	  reduction_memory_size);
  output_stream = NULL;

  // Initialize the ``previous record'' values to zero, which is the
  // assumed starting point for difference encoding purposes.
  previous_cycle_timestamp = 0;
  previous_instruction_timestamp = 0;
  previous_reference_timestamp = 0;
  previous_context_ID = 0;
  previous_page_ID = 0;

} // SAD_filter::SAD_filter  
// ===================================================================



// ===================================================================
SAD_filter::~SAD_filter () {

  // There is only work to do if the output stream was opened.
  if (output_stream != NULL) {

    // Flush the window completely, as all records still within it
    // have survived reduction.
    flush_window(window.end());

    // Emit an end-of-trace token and ensure that it succeeded.
    int result = gzprintf(output_stream, "%c\n", TAG_END_OF_TRACE);
    if (result == 0) {
      fprintf(stderr, "SAD_filter::~SAD_filter(): Print failure\n");
      exit(1);
    }

    // Close the output stream, ensuring that is succeeded.
    result = gzclose(output_stream);
    if (result != Z_OK) {
      fprintf(stderr, "SAD_filter::~SAD_filter(): Close failure\n");
      exit(1);
    }

  }

} // SAD_filter::~SAD_filter
// ===================================================================



// ===================================================================
void
SAD_filter::open_output () {

  // It is an error to open an already open output stream.
  if (output_stream != NULL) {
    fprintf(stderr, "SAD_filter::open_output(): Stream already open\n");
    exit(1);
  }
  
  // Attempt to open the output file.
  int fd = open(output_pathname,
		O_WRONLY | O_CREAT | O_EXCL, 
		S_IRUSR | S_IWUSR);

  // Ensure that it did open.
  if (fd == -1) {
    perror("SAD_filter()::open_output(): Failed file open");
    exit(1);
  }

  // Link the file descriptor to a compressed output stream.  Ask that
  // the stream be for (w)riting at the highest (9) compression level
  // accepting (f)iltered data.
  output_stream = gzdopen(fd, "w9f");

  // Ensure that the linking succeeded.
  if (output_stream == NULL) {
    fprintf(stderr,
	    "SAD_filter::open_output(): Failed compression open\n");
    exit(1);
  }

  // Since the task associated with this filter actually has
  // references to be processed, emit its task ID into the file that
  // holds such a list.
  fprintf(task_ID_list_stream, "%d\n", task_ID);

} // SAD_filter::open_output
// ===================================================================



// ===================================================================
void
SAD_filter::append_record (SAD_record* const record) {

  // If the output stream is not already open, then open it.
  if (output_stream == NULL) {
    open_output();
  }

  // Update the LRU ordering based on this reference.
  page_s evicted = lru_queue.touch(&record->page);

  // Did an eviction occur?
  if (evicted != lru_queue.no_evict) {

    // Yes, so move forward the back end of the window.
    move_forward(evicted);

  } else {

    // No eviction occurred.  Grab the number of references to this
    // page already
    int* references = &(number_of_references[record->page]);

    // Are there at least two references to this page already in the
    // run? (for a total of three together with the current one)  Note
    // that there may be more than two if an earlier record has been
    // protected from being dropped.
    if (*references >= 2) {

      // Yes.  Attempt to erase the middle reference.  Note that it
      // may not be erased if the middle reference is protected.
	erase_middle_reference(record, references);

    }
  }

  // Increment the number of references to this page within the
  // window, place the new record into the window, and mark it as the
  // last reference to this page in the window.
  number_of_references[record->page]++;
  window.push_back(*record);
  list<SAD_record>::iterator i = window.end();
  i--;
  last_reference_map[record->page] = i;

} // SAD_filter::append_record
// ===================================================================



// ===================================================================
void
SAD_filter::append_kernel_event_record (SAD_record* const record) {

  // Sanity check: Ensure that this _is_ a kernel event.
  if (record->tag != TAG_KERNEL_EVENT) {
    fprintf(stderr, "SAD_filter::append_kernel_event_record(): ");
    fprintf(stderr, "Non-kernel event passed\n");
    exit(1);
  }

  // If the output stream is not already open, then open it.
  if (output_stream == NULL) {
    open_output();
  }

  // Add this record to the window.
  window.push_back(*record);

} // SAD_filter::append_kernel_event_record
// ===================================================================



// ===================================================================
// If an eviction occurs, the window representing all pages in an LRU
// window of size less than queue_size needs to be adjusted.
void
SAD_filter::move_forward (page_s evicted) {

  // Grab the iterator to the last reference in the window to the
  // evicted page.  Then move to the following reference to provide a
  // stopping point for the flushing of records from the window.
  list<SAD_record>::iterator first_to_keep =
    last_reference_map[evicted];
  first_to_keep++;

  // Flush the oldest records from the window up to and including the
  // most recenct record that references the given evicted page.
  flush_window(first_to_keep);

} // SAD_filter::move_forward
// ===================================================================



// ===================================================================
void
SAD_filter::flush_window
(const list<SAD_record>::iterator first_to_keep) {

  // Move the front of the window forward until the new front is
  // reached, emitting those records that have survived reduction
  // while removing them from the window.
  list<SAD_record>::iterator i = window.begin();
  list<SAD_record>::iterator j = i;
  j++;
  while (i != first_to_keep) {

    // There is extra bookkeeping to do only for actual reference
    // records, and not for kernel event records.
    if ((*i).tag != TAG_KERNEL_EVENT) {

      // Update the count of records in the window.  If there are no
      // references left to a given page, erase its entry in the
      // reference count table.
      int* references = &(number_of_references[(*i).page]);
      (*references)--;
      if (*references == 0) {
	last_reference_map[(*i).page] = window.end();
	number_of_references.erase((*i).page);
      }

      // If the next record references the same page as this one, then
      // it must be the case that there are two records of references
      // to this page in the window, and that they are back-to-back.
      // The second record is superfluous -- it is not needed to
      // ensure that this page will be evicted at the correct moment
      // -- and therefore can be dropped.
      if ((*i).is_same_page(*j)) {
	(*i).transfer_dirtiness_from(*j);
	list<SAD_record>::iterator moribund = j;
	j++;
	window.erase(moribund);
      }

    } // Bookkeeping for real reference records

    // Emit this record that has survived reduction.
    emit(&(*i));

    // Delete the record and move to the next one.
    i++;
    j++;
    window.pop_front();

  } // Loop through records

} // SAD_filter::flush_window
// ===================================================================



// ===================================================================
// When we recognize a triple of the correct form we drop the middle
// reference.
void
SAD_filter::erase_middle_reference (SAD_record* const record,
				    int* const references) {

  // Grab a pointer to the last reference in the window for this
  // page.
  list<SAD_record>::iterator middle_record =
    last_reference_map[record->page];
    
  // Check the sanity of this middle reference.
  assert((*middle_record).page == record->page);

  // We can drop the middle record only if it is not protected.
  if (!(*middle_record).drop_protected) {

    // Update the current record with the dirtying property of the
    // dropped record.
    record->transfer_dirtiness_from(*middle_record);

    // Drop the middle record.
    window.erase(middle_record);

    // Temporarily set the middle reference for this page to null.
    // (It will be updated when the current record is inserted into
    // the window).
    last_reference_map[record->page] = window.end();

    // Decrement the counter of references to the given page within
    // the window.
    (*references)--;

  }

} // SAD_filter::erase_middle_reference
// ===================================================================



// ===================================================================
void
SAD_filter::emit (const SAD_record* const record) {

  // Calculate the differences for each timestamp (always positive).
  timestamp_t cycle_timestamp_difference =
    record->cycle_timestamp - previous_cycle_timestamp;
  timestamp_t instruction_timestamp_difference =
    record->instruction_timestamp - previous_instruction_timestamp;
  timestamp_t reference_timestamp_difference =
    record->reference_timestamp - previous_reference_timestamp;

  // Store the current timestamps for later.
  previous_cycle_timestamp = record->cycle_timestamp; 
  previous_instruction_timestamp = record->instruction_timestamp;
  previous_reference_timestamp = record->reference_timestamp;

  // Calculate the context ID to emit.  If it is the same as the
  // previous context ID, emit zero; otherwise, emit the new context
  // ID.
  context_ID_t context_ID_output = 0;
  if (record->page.context_ID != previous_context_ID) {
    context_ID_output = record->page.context_ID;
    previous_context_ID = record->page.context_ID;
  }

  // Calculate the page ID difference.  It can be positive or
  // negative, so we must determine both sign and magnitude.
  char page_ID_difference_sign;
  page_ID_t page_ID_difference_magnitude;
  if (record->page.page_ID < previous_page_ID) {

    page_ID_difference_sign = '-';
    page_ID_difference_magnitude = (previous_page_ID -
				    record->page.page_ID);

  } else {

    page_ID_difference_sign = '+';
    page_ID_difference_magnitude = (record->page.page_ID - 
				    previous_page_ID);

  }
  previous_page_ID = record->page.page_ID;

  // Print the difference encoded values to the compressed stream.
  int result = gzprintf(output_stream,
			"%c %qx %qx %qx %lx %c %lx\n",
			record->tag,
			cycle_timestamp_difference,
			instruction_timestamp_difference,
			reference_timestamp_difference,
			context_ID_output,
			page_ID_difference_sign,
			page_ID_difference_magnitude);

  // Ensure that the print succeeded.
  if (result == 0) {
    fprintf(stderr, "SAD_filter::emit(): Print failed\n");
    exit(1);
  }
} // emit
// ===================================================================
