/**
 * \file SAD_filter.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu> and Yannis
 * Smaragdakis <yannis@cc.gatech.edu>
 * \date January 2003
 * \brief The header for the SAD class which performs trace
 * reduction.
 *
 * The definition of a SAD filter, which reduces a reference stream
 * and emits the results.
 **/



// =================================================================
#if !defined (_SAD_FILTER_HH)
#define _SAD_FILTER_HH
// =================================================================



// =================================================================
#include <hash_map.h>
#include <list.h>
#include <zlib.h>
#include "types-and-constants.hh"
#include "LRU_queue.hh"
#include "SAD_record.hh"
// =================================================================



// =================================================================
// TYPES

/** \brief A type for the lookup table that maps pages to numbers of
 *         references within the reduction window. */
typedef hash_map<page_s,
		 int,
		 page_hash_s,
		 equal_pages_s> page_count_map_t;

/** \brief A type for the lookup table that maps pages to iterators
 *         used to point to the last reference in the window. */
typedef hash_map<page_s,
		 list<SAD_record>::iterator,
		 page_hash_s,
		 equal_pages_s> last_reference_map_t;
// =================================================================



// =================================================================
/**
 * \brief A SAD_filter performs lossy reduction on a reference stream.
 *
 * A SAD filter accepts a sequence of page references and eliminates
 * those references that are ``superfluous''.  Given a reduction
 * memory size, SAD eliminates references that affect the order and
 * timing of neither the fetching or evicting of a page into or from
 * an LRU-managed memory of that size.
 **/
class SAD_filter {

public:

  /**
   * \brief The constructor.
   * \param reduction_memory_size The reduction memory size, in
   *        pages.
   * \param output_directory_pathname The name of the file into which
   *        the reduced stream will be emitted.
   * \param init_task_ID The task ID whose references this object will
   *        filter.
   * \param init_task_ID_list_stream The file that holds the IDs of
   *        tasks that actually performed referenced.
   *
   * Prepare a SAD filter by creating an LRU queue of the appropriate
   * size and constructing the pathname of the file into which a
   * reduced reference trace will be written (if any).  An output
   * stream will be opened only once actual references are passed into
   * the filter, at which point this object will record its own task
   * ID into a list of those that actually processed references.
   **/
  SAD_filter (const unsigned int reduction_memory_size,
	      const char* const output_directory_pathname,
	      const task_ID_t init_task_ID,
	      FILE* const init_task_ID_list_stream);

  /**
   * \brief The destructor.
   *
   * Flush what remains in the SAD window and close the output file.
   **/
  ~SAD_filter ();

  /**
   * \brief Process the next record.
   * \param record The next record in the reference stream to
   *        process.
   *
   * Insert this next record into the window so that it can be reduced
   * and, if it survives reduction, emitted.
   **/
  void append_record (SAD_record* const record);

  /**
   * \brief Insert a kernel event into the reference stream.
   * \param record The next record in the reference event stream to
   *        process.
   *
   * Insert a kernel event record into the window.  Since kernel event
   * records are simply placeholders to indicate when kernel records
   * will exist in the kernel trace, they cannot be reduced, and do
   * not require filtration.  By inserting them, they will simply pass
   * through the filter and be emitted in order.
   **/
  void append_kernel_event_record (SAD_record* const record);

protected: // METHODS

  /**
   * \brief Open the compressed output file.
   *
   * Open the output file, this providing a compressed stream to which
   * records can be written.  Note that this method should be called
   * only once, upon the first addition of a record to the filter.
   **/
  void open_output ();

  /**
   * \brief Update the trailing end of the window following an
   *        eviction.
   * \param evicted The evicted page.
   *
   * If an eviction occurs, then the back end of the window must be
   * moved forward until no instances of the evicted page remain.
   * This shrinking of the window will force surviving records to be
   * emitted to the output file.
   **/
  void move_forward (page_s evicted);

  /**
   * \brief Emit the older records in the window to a given point.
   * \param first_to_keep An iterator to the first record in the
   *        window that should <b>not</b> be emitted.
   *
   * Emit and erase records from the window starting with the oldest
   * record and proceeded until the given iterator is reached.  Note
   * that the record to which the iterator points is <b>not itself</b>
   * emitted and erased.  If back-to-back records that reference the
   * same page are found during this process, the latter record will
   * be dropped.
   **/
  void flush_window (const list<SAD_record>::iterator first_to_keep);

  /**
   * \brief Output a record that has survived SAD reduction.
   * \param record The record to output.
   *
   * Given a record that has survived SAD reduction, difference encode
   * its values and output it to the compressed file stream.
   **/
  void emit (const SAD_record* const record);

  /**
   * \brief Eliminate a superfluous record.
   * \param record The current record whose referenced page forms the
   *        last in a triplet.
   * \param references A pointer to the number of references to the
   *        given page already within the window.
   *
   * A triplet of references exists in the window, completed by the
   * current reference.  Find the middle reference and drop its
   * record <b>if</b> it is not protected from elimination.  If it is
   * protected, do nothing.  After dropping the record, update the
   * number of records for that page in the window via the given
   * pointer to that counter.
   **/
  void erase_middle_reference (SAD_record* const record,
			       int* const references);

protected: // DATA MEMBERS

  /** \brief The task ID whose references this object is filtering. */
  const task_ID_t task_ID;

  /** \brief The pathname of the compressed output stream. */
  char output_pathname[BUFFER_SIZE];

  /** \brief The compressed output stream to which reduced records are
   *         written. */
  gzFile output_stream;

  /** \brief The file into which the IDs of tasks that perform
   *         references are recorded. */
  FILE* const task_ID_list_stream;

  /** \brief The LRU queue that determines which pages belong in the
   *         window. */
  LRU_queue lru_queue;

  /** \brief A list representing the current reduction window of
   *         records. */
  list<SAD_record> window;

  /** \brief The table that maps pages to a count of references within
   *         the window. */
  page_count_map_t number_of_references;

  /** \brief The table that maps pages to iterators that mark the last
   *         reference to that page in the window. */
  last_reference_map_t last_reference_map;

  /** \brief The cycle timestamp of the previously emitted record. */
  timestamp_t previous_cycle_timestamp;

  /** \brief The instruction timestamp of the previously emitted
   *         record. */
  timestamp_t previous_instruction_timestamp;

  /** \brief The reference timestamp of the previously emitted
   *         record. */
  timestamp_t previous_reference_timestamp;

  /** \brief The context ID of the previously emitted record. */
  context_ID_t previous_context_ID;

  /** \brief The page ID of the previous emitted record. */
  page_ID_t previous_page_ID;

}; // SAD_filter
// =================================================================



// =================================================================
#endif // _SAD_FILTER_HH
// =================================================================
