/**
 * \file SAD_record.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief The header for the SAD_record class.
 *
 * This file defines a SAD_record object, which holds the needed
 * information and operations about a trace record needed for
 * reduction with SAD.
 **/

// ===================================================================
// Avoid multiple inclusion.
#if !defined (_SAD_RECORD_HH)
#define _SAD_RECORD_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <list.h>
#include <set.h>
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
/**
 * \brief An object representing a single reference trace record.
 * We violate the usual rule about protecting object data, exposing it
 * as <b>public</b>, for efficiency.
 **/
class SAD_record {

public: // DATA MEMBERS

  tag_t tag; /**< The type of reference that was performed. */
  bool dirtying; /**< Whether the reference dirtied its page. */
  timestamp_t cycle_timestamp; /**< The cycle time at which the reference
				  occurred. */
  timestamp_t instruction_timestamp; /**< The instruction time of the
					reference. */
  timestamp_t reference_timestamp; /**< The reference time of the
				      reference. */
  page_s page; /**< The page that was referenced in this record. */
  bool drop_protected; /**< A flag that indicates whether this page
			  has been protected from elimination. */

public:

  /**
   * \brief The default constructor.
   * Initialize the fields of a new record to zero.
   **/
  SAD_record ();

  /**
   * \brief Determine if two records reference the same page.
   * \param other_record The other record to be compared to this one.
   * \return Whether this record on the other reference the same
   * page
   **/
  bool is_same_page (const SAD_record& other_record) const;

  /**
   * \brief Transfer the dirtiness attribute of another record to this
   *        one, where both reference the same page.
   * \param source_record The other record that referenced the
   *        same page.
   *
   * If the provided source record is for a dirtying reference to the
   * page, then this record will also become a dirtying reference.  If
   * the source record is not dirtying, then this one will retain its
   * original dirtying property.
   **/
  void transfer_dirtiness_from (const SAD_record& source_record);

};
// ===================================================================



// ===================================================================
typedef set<page_s, less<page_s> > set_of_pages;
// ===================================================================



// ===================================================================
#endif // _SAD_RECORD_HH
// ===================================================================
