/**
 * \file main.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date December 2002
 * \brief The code for page-level reduction reference handler.
 *
 * This file contains all of the code for page-level reduction
 * reference handler, which converts the binary output of the Laplace
 * raw reference trace into a SAD-reduced, difference encoded, text
 * reference trace format.
 */



// ===================================================================
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h> // Needed for strtoul()
#include "types-and-constants.hh"
#include "SAD_filter.hh"
#include <fcntl.h>
#include <sys/types.h>
// ===================================================================




// ===================================================================
/**
 * \brief Read a single record from the binary raw reference trace.
 * \param input_reference_trace_stream The stream for the binary raw
 *        reference trace.
 * \param binary_record A pointer to a binary record structure into
 *        which the results will be placed.
 *
 * Attempt a read the next record from the binary raw reference trace
 * stream into the provided structure.  If the read fails, abort
 * processing.
 **/
void
read_reference_record (FILE* const input_reference_trace_stream,
		       reference_record_s* const binary_record) {

  // Attempt to read the next record.
  size_t read_result = fread(binary_record,
			     sizeof(*binary_record),
			     1,
			     input_reference_trace_stream);

  // Was there a problem reading?
  if (read_result != 1) {

    // There was.  Differentiate between error types and then abort
    // processing.
    if (feof(stdin)) {
      fprintf(stderr, "Premature EOF\n");
    } else {
      perror("Failed in Reading binary record");
      fprintf (stderr, "fread returned %d\n", read_result);
    }

    exit(1);

  }
} // read_reference_record
// ===================================================================



// ===================================================================
/**
 * \brief Open a new filtered output stream for a task.
 * \param output_directory_pathname The filtered stream will be
 *        emitted to a new file in this directory.
 * \param reduction_memory_size The reduction memory size that the SAD
 *        filter will use.
 * \param task_ID_to_SAD_filter_map Once the filter is created, the
 *        table that will map the task to the filter.
 * \param task_ID The new task for whom a filter is being created.
 * \param task_ID_list_stream The file that stores the list of task
 *        IDs for which we observed references.
 * \return The new SAD filter.
 *
 * Create a new SAD filter, configuring it to emit its output to a new
 * file whose name is associated with the given task ID.  Record in
 * the given table the association between the given task ID and the
 * new filter.
 **/
SAD_filter*
open_filter
(const char* const output_directory_pathname,
 const unsigned int reduction_memory_size,
 task_ID_to_SAD_filter_map_t* const task_ID_to_SAD_filter_map,
 const task_ID_t task_ID,
 FILE* const task_ID_list_stream) {

  // Create the new SAD filter.
  SAD_filter* new_filter = new SAD_filter(reduction_memory_size,
					  output_directory_pathname,
					  task_ID,
					  task_ID_list_stream);

  // Establish the new mapping between task and filter.  First,
  // however, ensure that there isn't already a filter associated with
  // this task.
  task_ID_to_SAD_filter_map_t::iterator i =
    (*task_ID_to_SAD_filter_map).find(task_ID);
  if (i != (*task_ID_to_SAD_filter_map).end()) {
    fprintf(stderr, "open_filter(): Task already mapped to a filter\n");
    exit(1);
  }
  (*task_ID_to_SAD_filter_map)[task_ID] = new_filter;

  // Return the new filter to the caller.
  return new_filter;

} // open_filter
// ===================================================================



// ===================================================================
/**
 * \brief Close a filtered output stream associated with a particular
 *        task.
 * \param task_ID_to_SAD_filter_map The table that maps tasks to
 *        filters.
 * \param task_ID The task whose filtered output should be closed.
 *
 * Lookup the filter associated with the given task via the given
 * table.  Delete the filter (thus closing its stream) and erase the
 * mapping.
 **/
void
close_filter
(task_ID_to_SAD_filter_map_t* const task_ID_to_SAD_filter_map,
 const task_ID_t task_ID) {

  // Attempt to look up the filter based on the task ID.  If it
  // failed, then abort processing.
  task_ID_to_SAD_filter_map_t::iterator i =
    (*task_ID_to_SAD_filter_map).find(task_ID);
  if (i == (*task_ID_to_SAD_filter_map).end()) {
    fprintf(stderr, "close_filter(): No filter for given task\n");
    exit(1);
  }
      
  // Delete the filter (which will close the stream) and erase the
  // entry from the map.
  delete (*i).second;
  (*task_ID_to_SAD_filter_map).erase(i);

} // close_filter
// ===================================================================



// ===================================================================
/**
 * \brief Close all filtered output streams in a given table.
 * \param task_ID_to_SAD_filter_map The table mapping tasks to
 *        filters.
 *
 * Delete every filter (thus closing every output stream) found in
 * the given table, and erase all task->filter mappings.
 **/
void
close_all_filters
(task_ID_to_SAD_filter_map_t* const task_ID_to_SAD_filter_map) {

  // Traverse every entry in the table.
  task_ID_to_SAD_filter_map_t::iterator i =
    (*task_ID_to_SAD_filter_map).begin();
  while (i != (*task_ID_to_SAD_filter_map).end()) {

    // Delete the filter (thus closing the stream).  Advance the
    // iterator before erasing the mapping so that the iteration
    // works.
    delete (*i).second;
    task_ID_to_SAD_filter_map_t::iterator moribund = i;
    i++;
    (*task_ID_to_SAD_filter_map).erase(moribund);

  }

} // close_all_filters
// ===================================================================



// ===================================================================
/**
 * \brief Read a record from a kernel trace.
 * \param input_kernel_trace_stream The input stream for the kernel
 *        trace.
 * \param kernel_record The structure into which the record will be
 *        placed.
 * \param output_kernel_trace_stream The compressed stream to which
 *        the kernel trace will be written.
 *
 * Attempt to read a single record from the given kernel trace.  The
 * results are placed into the provided structure.  A copy of the
 * kernel record is also written to the output kernel trace stream:
 * Since reference handlers are designed to be used on-line, this
 * allows the machine simulator to pipe both its kernel and reference
 * traces to this utility and have <b>both</b> of them stored in
 * compressed form in the output directory.
 **/
void
read_kernel_record (FILE* const input_kernel_trace_stream,
		    gzFile const output_kernel_trace_stream,
		    kernel_record_s* const kernel_record) {

  // Attempt to read the next record of the kernel trace.
  char buffer[BUFFER_SIZE];

  char* read_result = fgets(buffer,
			    BUFFER_SIZE,
			    input_kernel_trace_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.
    if (feof(input_kernel_trace_stream)) {
      fprintf(stderr,
	      "read_kernel_record(): Premature EOF\n");
    } else {
      perror("read_kernel_record(): Read error\n");
    }
    exit(1);
  }


  // Write a copy of the record to the compressed output stream and
  // ensure that it succeeded.
  int write_result = gzputs(output_kernel_trace_stream, buffer);
  if (write_result == -1) {
    fprintf(stderr,
	    "read_kernel_record(): Failed kernel record write\n");
    exit(1);
  }

  // Parse the record type, which should be the first character of the
  // buffer.
  kernel_record->tag = buffer[0];

  // Parse the timestamp if this is not an end-of-trace record.
  if (kernel_record->tag != TAG_END_OF_TRACE) {

    int parse_result = sscanf(&buffer[2],
			      "%qx",
			      &kernel_record->timestamp);
    if (parse_result != 1) {
      fprintf(stderr, "read_kernel_record(): Failed timestamp parse\n");
      exit(1);
    }
  }

  // Only bother parsing the remainder of the record if it is one of
  // the record types that this utility must handle.
  if ((kernel_record->tag == TAG_SCHEDULE) ||
      (kernel_record->tag == TAG_FORK) ||
      (kernel_record->tag == TAG_EXIT)) {

    // Attempt to read the timestamp and task ID.  (Re-parse the
    // timestamp simple because we wouldn't know where to begin
    // parsing otherwise.)
    int parse_result = sscanf(&buffer[2],
			      "%qx %hd",
			      &kernel_record->timestamp,
			      &kernel_record->task_ID);

    // Ensure that the parse succeeded.
    if (parse_result != 2) {
      fprintf(stderr,
	      "read_kernel_record(): Failed remaining parse\n");
      exit(1);
    } // Unsuccessful parse
  } // Interesting record type

} // read_kernel_record
// ===================================================================



// ===================================================================
/**
 * \brief Process a kernel event.
 * \param kernel_record The record of a kernel event to be processed.
 * \param output_directory_pathname The directory into which filtered
 *        output files are placed.
 * \param reduction_memory_size The reduction memory size to be used
 *        if a new filter is created.
 * \param task_ID_to_SAD_filter_map The table that maps tasks to
 *        filtered output streams.
 * \param task_ID_list_stream The file into which we will store a list
 *        of task IDs that perform references.
 * \param scheduled_task_ID The ID of the currently scheduled user
 *        task, taken by reference to enable changes.
 * \param scheduled_filter The SAD filter for the currently scheduled
 *        user task, taken by reference to enable changes.
 *
 * Read the next kernel trace record and process it.  Specifically,
 * if the kernel record indicates that a new process has been created,
 * an existing process has been terminated, or an existing process has
 * become the current one (that is, scheduled), then adjust our data
 * accordingly.
 **/
void
process_kernel_record
(kernel_record_s* const kernel_record,
 const char* const output_directory_pathname,
 const unsigned int reduction_memory_size,
 task_ID_to_SAD_filter_map_t* const task_ID_to_SAD_filter_map,
 FILE* const task_ID_list_stream,
 task_ID_t* const scheduled_task_ID,
 SAD_filter** const scheduled_filter) {

  // Parse the kernel record based on its type.  Declare and
  // initialize some variables here to avoid compiler warnings.
  task_ID_to_SAD_filter_map_t::iterator i =
    (*task_ID_to_SAD_filter_map).end();
  SAD_filter* new_filter = 0;
  switch(kernel_record->tag) {

  case TAG_SCHEDULE:

    // Update the current task and filter.
    *scheduled_task_ID = kernel_record->task_ID;
    i = (*task_ID_to_SAD_filter_map).find(kernel_record->task_ID);
    if (i == (*task_ID_to_SAD_filter_map).end()) {
      fprintf(stderr,
	      "process_kernel_record(): No entry for scheduled task\n");
      exit(1);
    }
    *scheduled_filter = (*i).second;
    break;

  case TAG_FORK:

    // Open a new filter associated with this new task ID and add it
    // to the map.
    new_filter = open_filter(output_directory_pathname,
			     reduction_memory_size,
			     task_ID_to_SAD_filter_map,
			     kernel_record->task_ID,
			     task_ID_list_stream);
    (*task_ID_to_SAD_filter_map)[kernel_record->task_ID] = new_filter;
    break;

  case TAG_EXIT:

    // Close a filter associated with this now-defunct task ID.
    close_filter(task_ID_to_SAD_filter_map, kernel_record->task_ID);

    // If we just closed the current task, make the filter
    // inaccessible.
    *scheduled_task_ID = 0;
    *scheduled_filter = 0;
    break;

  case TAG_END_OF_TRACE:

    // The kernel trace cannot end before the reference trace.
    fprintf(stderr,
	    "process_kernel_record(): Premature kernel end-of-trace\n");
    exit(1);
    break;

  default:

    // Not an event type of concern to us, so ignore it.
    ;

  }

} // process_kernel_record
// ===================================================================



// ===================================================================
/**
 * \brief Synchronize the kernel and reference traces.
 * \param input_kernel_trace_stream The input file stream for the
 *        kernel trace.
 * \param output_kernel_trace_stream The compressed stream to which
 *        the kernel trace will be written.
 * \param last_kernel_event_time The time of the last kernel event
 *        preceeding the first reference event.
 * \param output_directory_pathname The directory into which filtered
 *        output files are placed.
 * \param reduction_memory_size The reduction memory size to be used
 *        if a new filter is created.
 * \param task_ID_to_SAD_filter_map The table that maps tasks to
 *        filtered output streams.
 * \param task_ID_list_stream The file into which we will store a list
 *        of task IDs that perform references.
 * \param scheduled_task_ID The ID of the currently scheduled user
 *        task, taken by reference to enable changes.
 * \param scheduled_filter The SAD filter for the currently scheduled
 *        user task, taken by reference to enable changes.
 *
 * Read and process all kernel records up to and including the last
 * kernel event that preceeded real reference event, thus bringing the
 * two traces into synchrony.  Set the current task ID and filter
 * based on the last kernel scheduling event.
 **/
void
synchronize
(FILE* const input_kernel_trace_stream,
 gzFile const output_kernel_trace_stream,
 const timestamp_t last_kernel_event_time,
 const char* const output_directory_pathname,
 const unsigned int reduction_memory_size,
 task_ID_to_SAD_filter_map_t* const task_ID_to_SAD_filter_map,
 FILE* const task_ID_list_stream,
 task_ID_t* const scheduled_task_ID,
 SAD_filter** const scheduled_filter) {

  // If the last kernel event time is 0, that means that reference
  // tracing began during system startup, _before_ the kernel was
  // loaded and any kernel events could be recorded.  Therefore, we
  // can skip synchronization.
  if (last_kernel_event_time == 0) {
    return;
  }

  // Repeatedly read and process kernel events until we read the last
  // one preceeding real references.
  kernel_record_s kernel_record;
  do {

    // Read the next kernel record.
    read_kernel_record(input_kernel_trace_stream,
		       output_kernel_trace_stream,
		       &kernel_record);

    // Sanity check:  We shouldn't go too far in the kernel trace.
    if (kernel_record.timestamp > last_kernel_event_time) {
      fprintf(stderr,
	      "synchronize(): Went too far in kernel trace\n");
      exit(1);
    }

    // Process the record.
    process_kernel_record(&kernel_record,
			  output_directory_pathname,
			  reduction_memory_size,
			  task_ID_to_SAD_filter_map,
			  task_ID_list_stream,
			  scheduled_task_ID,
			  scheduled_filter);

    // If this was the last record, then the loop will exit here.
  } while (kernel_record.timestamp != last_kernel_event_time);

} // synchronize
// ===================================================================



// ===================================================================
/**
 * \brief Read and write kernel records until the kernel trace has
 *        been completely processed.
 * \param input_kernel_trace_stream The stream from which kernel
 *        records are read.
 * \param output_kernel_trace_stream The compressed stream to which
 *        kernel records will be written.
 *
 * Simply read kernel records from the input stream and write them to
 * the compressed output stream.  This function should be called after
 * all of the reference records have been processed, and this utility
 * needs only to finish emitting the kernel trace to its compressed
 * file.
 **/
void
complete_kernel_trace (FILE* const input_kernel_trace_stream,
		       gzFile const output_kernel_trace_stream) {

  // Read and write kernel records until the end-of-trace is reached.
  char buffer[BUFFER_SIZE];
  do {

    // Attempt to read the next kernel record.
    char* read_result = fgets(buffer,
			      BUFFER_SIZE,
			      input_kernel_trace_stream);

    // Did the read fail?
    if (read_result == NULL) {

      // Distinguish between an EOF and other errors, but abort
      // processing in either case.
      if (feof(input_kernel_trace_stream)) {
	fprintf(stderr,
		"complete_kernel_trace(): Premature EOF\n");
      } else {
	perror("complete_kernel_trace(): Read error\n");
      }
      exit(1);
    }

    // Attempt to write the next kernel record and check for success.
    int write_result = gzputs(output_kernel_trace_stream, buffer);
    if (write_result == -1) {
      fprintf(stderr, "complete_kernel_trace(): Failed write\n");
      exit(1);
    }

  } while (buffer[0] != TAG_END_OF_TRACE);

} // complete_kernel_trace
// ===================================================================



// ===================================================================
void
determine_protection (SAD_record* const reduction_record,
		      SAD_filter* const kernel_filter,
		      SAD_filter* const scheduled_filter,
		      SAD_filter** const active_filter) {

  // Determine what to do by reference type.
  switch (reduction_record->tag) {

  case TAG_USER_READ:
  case TAG_USER_WRITE:
  case TAG_USER_INSTRUCTION_FETCH:

    // This is a user-level reference.  Is this a change?
    if (*active_filter != scheduled_filter) {

      // It is a change.  Update the active filter and protect this
      // reference so that it will be present in the reduced trace,
      // and thus provide needed timing information.
      *active_filter = scheduled_filter;
      reduction_record->drop_protected = true;

    } else {

      // It is not a change.  This record may be dropped.
      reduction_record->drop_protected = false;

    }
    break;

  case TAG_KERNEL_READ:
  case TAG_KERNEL_WRITE:
  case TAG_KERNEL_INSTRUCTION_FETCH:

    // This is a kernel-level reference.  Is this a change?
    if (*active_filter != kernel_filter) {

      // It is a change.  Update the active filter and protect this
      // reference so that it will be present in the reduced trace,
      // and thus provide needed timing information.
      *active_filter = kernel_filter;
      reduction_record->drop_protected = true;

    } else {

      // It is not a change.  This record may be dropped.
      reduction_record->drop_protected = false;

    }
    break;

  default:

    fprintf(stderr,
	    "determine_protection(): Invalid tag %c(%hd)\n",
	    reduction_record->tag,
	    (unsigned short int)reduction_record->tag);
    exit(1);

  }

} // determine_protection
// ===================================================================



// ===================================================================
/**
 * \brief Divide and reduce the input references on a per-task basis.
 * \param input_reference_trace_stream The stream for the binary raw
 *        reference trace to read.
 * \param input_kernel_trace_stream The stream for the kernel trace to
 *        read.
 * \param output_kernel_trace_stream The compressed stream to which
 *        the kernel trace will be written.
 * \param task_ID_list_stream The file into which we will store a list
 *        of task IDs that perform references.
 * \param output_directory_pathname The directory in which the
 *        per-task output files will be created.
 * \param reduction_memory_size The reduction memory size that each
 *        SAD filter will use.
 *
 * Read the reference and kernel traces, separating the reference
 * stream into multiple streams on a per-task basis.  Each task's
 * reference stream will be reduced with a SAD filter, difference
 * encoded, compressed, and written to its own file.  For each task
 * that actually has references in the reference trace, emit its task
 * ID to a file that will list all such task IDs.
 **/
void
process_traces (FILE* const input_reference_trace_stream,
		FILE* const input_kernel_trace_stream,
		gzFile const output_kernel_trace_stream,
		FILE* const task_ID_list_stream,
		const char* const output_directory_pathname,
		const unsigned int reduction_memory_size) {

  // Track the passage of instruction and reference time as references
  // are read from the input.  Initialize their clocks here.
  timestamp_t instruction_clock = 0;
  timestamp_t reference_clock = 0;

  // Create the table that maps task IDs to SAD filters.
  task_ID_to_SAD_filter_map_t* task_ID_to_SAD_filter_map =
    new task_ID_to_SAD_filter_map_t;

  // Assume a single kernel task (with an ID of 0).  Create its SAD
  // filter which will exist until processing is complete.
  const task_ID_t kernel_task_ID = 0;
  SAD_filter* kernel_filter =
    open_filter(output_directory_pathname,
                        reduction_memory_size,
                        task_ID_to_SAD_filter_map,
                        kernel_task_ID,
                        task_ID_list_stream);

  // The scheduled filter is the filter for the user task that is
  // currently scheduled.  The active filter can either be the kernel
  // filter or this filter.
  task_ID_t scheduled_task_ID = 0;
  SAD_filter* scheduled_filter = NULL;

  // Read kernel trace records until we encounter a record that
  // indicates that reference tracing has begun (a TAG_MODE_CHANGED
  // record).
  kernel_record_s kernel_record;
  do {

    // Read the next kernel record.
    read_kernel_record(input_kernel_trace_stream,
		       output_kernel_trace_stream,
		       &kernel_record);


    // If this was the record that indicates that reference tracing has
    // begun, then the loop will exit here.
    if (kernel_record.tag == TAG_MODE_CHANGED) {
      break;
    }
    else { // Process the record
      process_kernel_record(&kernel_record,
			    output_directory_pathname,
			    reduction_memory_size,
			    task_ID_to_SAD_filter_map,
			    task_ID_list_stream,
			    &scheduled_task_ID,
			    &scheduled_filter);
    }
  } while (true);

  // Now we switch mode. We will now read reference records, and only
  // read a kernel record if the reference record indicates that we
  // should.

  // Track the active task (via its filter), assuming nothing initally
  // about whether the kernel or user task will be active first.
  SAD_filter* active_filter = NULL;

  // Process records from both traces until the end of the reference
  // trace is encountered.
  reference_record_s binary_record;
  SAD_record reduction_record;
  read_reference_record(input_reference_trace_stream, &binary_record);
  while (binary_record.tag != TAG_END_OF_TRACE) {

    // Transfer the basic data from the binary record to a SAD record.
    reduction_record.tag = binary_record.tag;
    reduction_record.cycle_timestamp = binary_record.timestamp;
    reduction_record.instruction_timestamp = instruction_clock;
    reduction_record.reference_timestamp = reference_clock;
    reduction_record.page.context_ID = binary_record.context_ID;
    reduction_record.page.page_ID =
      binary_record.virtual_address >> PAGE_SIZE_ORDER;

    // Determine how to transfer the remaining data and what actions
    // to take based on the type of reference.
    switch (binary_record.tag) {

    case TAG_USER_INSTRUCTION_FETCH:

      if (!IGNORE_INSTRUCTION_FETCHES) {

	// Convert this to a read operation, submit the record for
	// filtration, and advance time appropriately.
	reduction_record.tag = TAG_USER_READ;
	reduction_record.dirtying = false;
	determine_protection(&reduction_record,
			     kernel_filter,
			     scheduled_filter,
			     &active_filter);

	if (active_filter == NULL)
	  printf ("active_filter is NULL\n");
	active_filter->append_record(&reduction_record);
	reference_clock++;

      }

      instruction_clock++;
      break;

    case TAG_KERNEL_INSTRUCTION_FETCH:

      if (!IGNORE_INSTRUCTION_FETCHES) {

	// Convert this to a read operation, submit the record for
	// filtration, and advance time appropriately.
	reduction_record.tag = TAG_KERNEL_READ;
	reduction_record.dirtying = false;
	determine_protection(&reduction_record,
			     kernel_filter,
			     scheduled_filter,
			     &active_filter);
	active_filter->append_record(&reduction_record);
	reference_clock++;

      }

      instruction_clock++;
      break;

    case TAG_USER_READ:
    case TAG_KERNEL_READ:

      // Submit the record for filtration and advance time
      // appropriately.
      reduction_record.dirtying = false;
      determine_protection(&reduction_record,
			   kernel_filter,
			   scheduled_filter,
			   &active_filter);
      
      active_filter->append_record(&reduction_record);
      reference_clock++;
      break;

    case TAG_USER_WRITE:
    case TAG_KERNEL_WRITE:

      // Submit the record for filtration and advance time
      // appropriately.
      reduction_record.dirtying = true;
      determine_protection(&reduction_record,
			   kernel_filter,
			   scheduled_filter,
			   &active_filter);

      active_filter->append_record(&reduction_record);
      reference_clock++;
      break;

    case TAG_KERNEL_EVENT:

      // Read the next kernel record.
      read_kernel_record(input_kernel_trace_stream,
			 output_kernel_trace_stream,
			 &kernel_record);

      // Ensure that it represents a kernel event that occurs at the
      // expected time.
      if (kernel_record.timestamp != binary_record.timestamp) {
	fprintf(stderr,
		"process_traces(): Kernel event time mismatch ");
	fprintf(stderr,
		"<kernel = %qx vs. reference = %qx>\n",
		kernel_record.timestamp,
		binary_record.timestamp);
	exit(1);
      }

      // Process the kernel event, potentially creating, destroying,
      // or making current some task.
      process_kernel_record(&kernel_record,
			    output_directory_pathname,
			    reduction_memory_size,
			    task_ID_to_SAD_filter_map,
			    task_ID_list_stream,
			    &scheduled_task_ID,
			    &scheduled_filter);

      // Pass the record to the kernel's filter, which records all
      // kernel events.
      kernel_filter->append_kernel_event_record(&reduction_record);
      break;

    case TAG_END_OF_TRACE:

      // This case should never arise, as this record should have
      // exited the loop.
      fprintf(stderr,
	      "process_traces(): TAG_END_OF_TRACE didn't exit loop\n");
      exit(1);
      break;

    default:

      fprintf(stderr,
	      "process_traces(): Unknown reference tag %c(%hd)\n",
	      binary_record.tag,
	      (unsigned short int)binary_record.tag);
      exit(1);

    }

    // Read the next binary record.
    read_reference_record(input_reference_trace_stream, &binary_record);

  }

  // Close all remaining SAD filters and delete the map.
  close_all_filters(task_ID_to_SAD_filter_map);
  delete task_ID_to_SAD_filter_map;
  task_ID_to_SAD_filter_map = 0;

  // Finish reading the kernel trace.
  complete_kernel_trace(input_kernel_trace_stream,
			output_kernel_trace_stream);

} // process_traces
// ===================================================================





// ===================================================================
/**
 * \brief Create a file that contains the reduction memory size.
 * \param output_directory_pathname The directory in which this file
 *        should be created.
 * \param reduction_memory_size The reduction memory size to be
 *        written into the file.
 *
 * So that a consumer of this utility's output can easily later
 * determine the reduction memory size used and reconstruct the
 * filenames of the per-task reduced traces, record the reduction
 * memory size in its own file.
 **/
void
emit_reduction_memory_size (const char* const output_directory_pathname,
			    const unsigned int reduction_memory_size) {

  // Attempt to open the file, ensure that it succeeds.
  char reduction_memory_size_pathname[BUFFER_SIZE];
  sprintf(reduction_memory_size_pathname,
	  "%s/%s",
	  output_directory_pathname,
	  REDUCTION_MEMORY_SIZE_FILENAME);
  FILE* reduction_memory_size_stream =
    fopen(reduction_memory_size_pathname, "w");
  if (reduction_memory_size_stream == NULL) {
    perror("Failed to open reduction memory size file");
    exit(1);
  }

  // Print the value into the file.
  fprintf(reduction_memory_size_stream, "%d\n", reduction_memory_size);

  // Close the file, ensuring success.
  int close_result = fclose(reduction_memory_size_stream);
  if (close_result == EOF) {
    perror("Failed to close reduction memory size file");
    exit(1);
  }

} // emit_reduction_memory_size
// ===================================================================



// ===================================================================
/**
 * \brief The entry point.
 * 
 * \param argc The number of arguments given at the command line.
 * \param argv The vector of argument strings passed at the command
 *             line.
 * \return An execution result code.
 * 
 * Parse the command line arguments, open the input streams, and call
 * for the processing to occur.
 **/
int
main (const unsigned int argc, const char** const argv) {

  // Were the correct number of arguments passed?
  if (argc != 5) {

    // No.  Emit usage and exit.
    fprintf(stderr, "Usage: %s\n", argv[0]);
    fprintf(stderr, "       <reference trace pathname>\n");
    fprintf(stderr, "       <kernel trace pathname>\n");
    fprintf(stderr, "       <output directory pathname>\n");
    fprintf(stderr, "       <reduction memory size>\n");
    return 1;

  }

  // Parse and name the arguments.
  const char* const input_reference_trace_pathname = argv[1];
  const char* const input_kernel_trace_pathname = argv[2];
  const char* const output_directory_pathname = argv[3];
  const unsigned long int reduction_memory_size = strtoul(argv[4],
							  NULL,
							  10);

  // Ensure that the reduction memory size was correctly parsed.
  if (errno != 0) {
    perror("Failed to parse reduction memory size");
    exit(1);
  }

  // Attempt to open the input reference and kernel traces and ensure
  // that they opened successfully.
  FILE* input_kernel_trace_stream =
  fopen(input_kernel_trace_pathname, "r");
  if (input_kernel_trace_stream == NULL) {
    perror("Failed to open input kernel trace");
    exit(1);
  }

  FILE* input_reference_trace_stream =
    fopen(input_reference_trace_pathname, "r");
  if (input_reference_trace_stream == NULL) {
    perror("Failed to open input reference trace");
    exit(1);
  }

  // Attempt to open the output kernel trace and ensure that it
  // succeeded.
  char output_kernel_trace_pathname[BUFFER_SIZE];
  sprintf(output_kernel_trace_pathname,
	  "%s/%s",
	  output_directory_pathname,
	  OUTPUT_KERNEL_TRACE_FILENAME);
  gzFile output_kernel_trace_stream =
    gzopen(output_kernel_trace_pathname, "w");
  if (output_kernel_trace_stream == NULL) {
    perror("Failed to open output kernel trace");
    exit(1);
  }

  // Attempt to open the task-ID list file and ensure that it
  // succeeded.
  char task_ID_list_pathname[BUFFER_SIZE];
  sprintf(task_ID_list_pathname,
	 "%s/%s",
	  output_directory_pathname,
	  TASK_ID_LIST_FILENAME);
  FILE* task_ID_list_stream = fopen(task_ID_list_pathname, "w");
  if (task_ID_list_stream == NULL) {
    perror("Failed to open task ID list file");
    exit(1);
  }


  // Emit a file that will indicate the reduction memory size used.
  emit_reduction_memory_size(output_directory_pathname,
			     reduction_memory_size);

  // Handle the references!
  process_traces(input_reference_trace_stream,
		 input_kernel_trace_stream,
		 output_kernel_trace_stream,
		 task_ID_list_stream,
		 output_directory_pathname,
		 reduction_memory_size);

  // Attempt to close the input traces and output kernel trace and
  // ensure that they closed successfully.
  int close_result = fclose(input_reference_trace_stream);
  if (close_result != 0) {
    perror("Failed to close input reference trace");
    exit(1);
  }
  close_result = fclose(input_kernel_trace_stream);
  if (close_result != 0) {
    perror("Failed to close input kernel trace");
    exit(1);
  }
  close_result = gzclose(output_kernel_trace_stream);
  if (close_result != Z_OK) {
    perror("Failed to close output kernel trace");
    exit(1);
  }

  // Exit successfully.
  return 0;

} // main
// ===================================================================
