/**
 * \file types-and-constants.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief The types and global constants used in many modules.
 *
 * This file defines a number of types used by many of the modules in
 * the page-level reduction reference handler.  It also declares a
 * number of useful constants.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_TYPES_AND_CONSTANTS_HH)
#define _TYPES_AND_CONSTANTS_HH
// ===================================================================



// =================================================================
// INCLUDES

#include <hash_map.h>
// =================================================================



// =================================================================
// FORWARD DECLARATIONS

class SAD_filter;
// =================================================================



// =================================================================
// TYPES
// Definitions of types for each field in the raw reference trace.
// This is taken from tracer.cc in Laplace itself.

/** \brief The type for the tag that indicates the type of a
 *  reference. */
typedef char tag_t;

/** \brief The type for the cycle counter timestamp. */
typedef unsigned long long int timestamp_t;

/** \brief The type for length values. */
typedef unsigned char length_t;

/**
 * \brief The type for the address space identifier.
 * The page number holding the base level of the page table.
 **/
typedef unsigned long int context_ID_t;

/** \brief The type for virtual addresses. */
typedef unsigned long int address_t;

/** \brief The type for virtual page numbers. */
typedef unsigned long int page_ID_t;

/** \brief The type for task identifiers. */
typedef unsigned short int task_ID_t;

/** \brief A record structure in which a reference record will be
 *         passed and stored. */
struct reference_record_s {

  tag_t tag;
  timestamp_t timestamp;
  length_t length;
  context_ID_t context_ID;
  address_t virtual_address;

}; // reference_record_s

/** \brief A record structure in which a kernel record will be passed
 *         and store. */
struct kernel_record_s {

  tag_t tag;
  timestamp_t timestamp;
  task_ID_t task_ID;

}; // kernel_record_s

/** \brief Hold the basic information that represents a page. */
struct page_s {

  context_ID_t context_ID; /**< The context in which the virtual
				    page exists. */
  page_ID_t page_ID; /**< The page's number within its
			      context. */

  /**
   * \brief The default constructor.
   * Set up a page with default values (0 context and page IDs).
   **/
  page_s () : context_ID(0), page_ID(0) {}

  /**
   * \brief The initializing constructor.
   * \param init_context_ID The context in which the new virtual page
   *        exists.
   * \param init_page_ID The new page's number within its context.
   *
   * Set up a page with the given identity.
   **/
  page_s (const context_ID_t init_context_ID,
	  const page_ID_t init_page_ID) :
    context_ID(init_context_ID),
    page_ID(init_page_ID) {}

  /**
   * \brief Compare two pages for equal identity.
   * \return Whether the pages have the same context and page IDs.
   **/
  friend bool operator== (const page_s& x, const page_s& y) {

    return ((x.context_ID == y.context_ID) && (x.page_ID == y.page_ID));

  }

  /**
   * \brief Compare two pages for differing identity.
   * \return Whether the pages have different context and page IDs.
   **/
  friend bool operator!= (const page_s& x, const page_s& y) {

    return ((x.context_ID != y.context_ID) || (x.page_ID != y.page_ID));

  }

}; // page_s

/** \brief A structure for comparing task IDs used by tables indexed
 * by task ID. */
struct equal_task_ID_s {
  bool operator() (const task_ID_t x, const task_ID_t y) {
    return (x == y);
  }
};

/** \brief A table type that maps task IDs to SAD filters. */
typedef hash_map<task_ID_t,
		 SAD_filter*,
		 hash<task_ID_t>,
		 equal_task_ID_s> task_ID_to_SAD_filter_map_t;

/** \brief A structure that exists to provide a method of comparison
 *         for the lookup table of pages. */
struct equal_pages_s {
  bool operator() (const page_s& x, const page_s& y) const {
    return (x == y);
  }
};

/** \brief A special hash function for hashing pages, which are not
 *         simple scalar values. */
struct page_hash_s {

  size_t operator() (const page_s& page) const {

    // Treat the value as a sequence of bytes, forming a hash from a
    // calculation borrowed for hashing strings.
    unsigned long int h = 0;
    char* s = (char*)&page;
    for (unsigned int i = 0; i < sizeof(page_s); i++) {
      h = (5 * h) + *s;
      s++;
    }

    return (size_t)h;
  }
};
// =================================================================



// =================================================================
// CONSTANTS

/** \brief The filename of the compressed kernel trace to be
 *         output. */
#define OUTPUT_KERNEL_TRACE_FILENAME "kernel-trace.gz"

/** \brief The filename of the task ID list. */
#define TASK_ID_LIST_FILENAME "task-ID-list"

/** \brief The filename of the reduction memory size file. */
#define REDUCTION_MEMORY_SIZE_FILENAME "SAD-reduction-memory-size"

/** \brief Whether to ignore instruction fetches. */
#define IGNORE_INSTRUCTION_FETCHES true

/**
 * \brief The number of low-order offset bits to each virtual address.
 * Given a page size n, the log_{2}(n), a.k.a. the number of lower
 * bits of an address to eliminate to create a page number.
 **/
#define PAGE_SIZE_ORDER 12

/** \brief The size that should be used for string buffers. */
#define BUFFER_SIZE 256

/** \brief The number of bits in a byte. */
#define BITS_PER_BYTE 8

/** \brief The tag indicating a user-level read (load) operation. */
#define TAG_USER_READ                'r'

/** \brief The tag indicating a kernel-level read (load) operation. */
#define TAG_KERNEL_READ              'R'

/** \brief The tag indicating a user-level write (store) operation. */
#define TAG_USER_WRITE               'w'

/** \brief The tag indicating a kernel-level write (store)
 *         operation. */
#define TAG_KERNEL_WRITE             'W'

/** \brief The tag indicating a user-level instruction fetch. */
#define TAG_USER_INSTRUCTION_FETCH   'i'

/** \brief The tag indicating reference tracing has begun. */
#define TAG_MODE_CHANGED    	     'H'

/** \brief The tag indicating a kernel-level instruction fetch. */
#define TAG_KERNEL_INSTRUCTION_FETCH 'I'

/** \brief The tag indicating a the scheduling of a task. */
#define TAG_SCHEDULE                 'S'

/** \brief The tag indicating the creation of a new task. */
#define TAG_FORK                     'F'

/** \brief The tag indicating the termination of a task. */
#define TAG_EXIT                     'T'

/**
 * \brief The tag indicating a kernel event.
 * A kernel event implies that there is a record that should be read
 * from the kernel trace in order to determine the details of the
 * event.  This tag allows there to be a placeholder in the reference
 * trace.
 **/
#define TAG_KERNEL_EVENT             'K'

/** \brief The tag indicating the end of the trace. */
#define TAG_END_OF_TRACE             'e'
// =================================================================



// =================================================================
#endif // _TYPES_AND_CONSTANTS_HH
// =================================================================
